@@ -1,5 +1,5 @@
 /* This file is part of Ganv.
- * Copyright 2007-2013 David Robillard <http://drobilla.net>
+ * Copyright 2007-2016 David Robillard <http://drobilla.net>
  *
  * Ganv is free software: you can redistribute it and/or modify it under the
  * terms of the GNU General Public License as published by the Free Software
@@ -26,9 +26,9 @@
 #include "./color.h"
 #include "./gettext.h"
 
+#include "color.h"
 #include "ganv-private.h"
 
-// TODO: Very sloppy clipping for arrow heads
 #define ARROW_DEPTH   32
 #define ARROW_BREADTH 32
 
@@ -47,6 +47,7 @@ enum {
 	PROP_DASH_LENGTH,
 	PROP_DASH_OFFSET,
 	PROP_COLOR,
+	PROP_CONSTRAINING,
 	PROP_CURVED,
 	PROP_ARROWHEAD,
 	PROP_SELECTED,
@@ -54,15 +55,15 @@ enum {
 	PROP_GHOST
 };
 
-G_DEFINE_TYPE(GanvEdge, ganv_edge, GANV_TYPE_ITEM)
+G_DEFINE_TYPE_WITH_CODE(GanvEdge, ganv_edge, GANV_TYPE_ITEM,
+                        G_ADD_PRIVATE(GanvEdge))
 
 static GanvItemClass* parent_class;
 
 static void
 ganv_edge_init(GanvEdge* edge)
 {
-	GanvEdgeImpl* impl = G_TYPE_INSTANCE_GET_PRIVATE(
-		edge, GANV_TYPE_EDGE, GanvEdgeImpl);
+	GanvEdgePrivate* impl = ganv_edge_get_instance_private(edge);
 
 	edge->impl = impl;
 
@@ -72,13 +73,14 @@ ganv_edge_init(GanvEdge* edge)
 	memset(&impl->coords, '\0', sizeof(GanvEdgeCoords));
 	impl->coords.width         = 2.0;
 	impl->coords.handle_radius = 4.0;
+	impl->coords.constraining  = TRUE;
 	impl->coords.curved        = FALSE;
 	impl->coords.arrowhead     = FALSE;
 
 	impl->old_coords  = impl->coords;
 	impl->dash_length = 0.0;
 	impl->dash_offset = 0.0;
-	impl->color       = 0xA0A0A0FF;
+	impl->color       = 0;
 }
 
 static void
@@ -108,23 +110,39 @@ ganv_edge_set_property(GObject*      object,
 	g_return_if_fail(object != NULL);
 	g_return_if_fail(GANV_IS_EDGE(object));
 
-	GanvEdge*       edge   = GANV_EDGE(object);
-	GanvEdgeImpl*   impl   = edge->impl;
-	GanvEdgeCoords* coords = &impl->coords;
+	GanvItem*        item   = GANV_ITEM(object);
+	GanvEdge*        edge   = GANV_EDGE(object);
+	GanvEdgePrivate* impl   = edge->impl;
+	GanvEdgeCoords*  coords = &impl->coords;
 
 	switch (prop_id) {
-		SET_CASE(TAIL, object, impl->tail);
-		SET_CASE(HEAD, object, impl->head);
 		SET_CASE(WIDTH, double, coords->width);
 		SET_CASE(HANDLE_RADIUS, double, coords->handle_radius);
 		SET_CASE(DASH_LENGTH, double, impl->dash_length);
 		SET_CASE(DASH_OFFSET, double, impl->dash_offset);
 		SET_CASE(COLOR, uint, impl->color);
+		SET_CASE(CONSTRAINING, boolean, impl->coords.constraining);
 		SET_CASE(CURVED, boolean, impl->coords.curved);
 		SET_CASE(ARROWHEAD, boolean, impl->coords.arrowhead);
 		SET_CASE(SELECTED, boolean, impl->selected);
 		SET_CASE(HIGHLIGHTED, boolean, impl->highlighted);
 		SET_CASE(GHOST, boolean, impl->ghost);
+	case PROP_TAIL: {
+		const gobject tmp = g_value_get_object(value);
+		if (impl->tail != tmp) {
+			impl->tail = GANV_NODE(tmp);
+			ganv_item_request_update(item);
+		}
+		break;
+	}
+	case PROP_HEAD: {
+		const gobject tmp = g_value_get_object(value);
+		if (impl->head != tmp) {
+			impl->head = GANV_NODE(tmp);
+			ganv_item_request_update(item);
+		}
+		break;
+	}
 	default:
 		G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
 		break;
@@ -140,8 +158,8 @@ ganv_edge_get_property(GObject*    object,
 	g_return_if_fail(object != NULL);
 	g_return_if_fail(GANV_IS_EDGE(object));
 
-	GanvEdge*     edge = GANV_EDGE(object);
-	GanvEdgeImpl* impl = edge->impl;
+	GanvEdge*        edge = GANV_EDGE(object);
+	GanvEdgePrivate* impl = edge->impl;
 
 	switch (prop_id) {
 		GET_CASE(TAIL, object, impl->tail);
@@ -151,6 +169,7 @@ ganv_edge_get_property(GObject*    object,
 		GET_CASE(DASH_LENGTH, double, impl->dash_length);
 		GET_CASE(DASH_OFFSET, double, impl->dash_offset);
 		GET_CASE(COLOR, uint, impl->color);
+		GET_CASE(CONSTRAINING, boolean, impl->coords.constraining);
 		GET_CASE(CURVED, boolean, impl->coords.curved);
 		GET_CASE(ARROWHEAD, boolean, impl->coords.arrowhead);
 		GET_CASE(SELECTED, boolean, impl->selected);
@@ -231,10 +250,10 @@ ganv_edge_bounds(GanvItem* item,
                  double* x1, double* y1,
                  double* x2, double* y2)
 {
-	GanvEdge*       edge   = GANV_EDGE(item);
-	GanvEdgeImpl*   impl   = edge->impl;
-	GanvEdgeCoords* coords = &impl->coords;
-	const double    w      = coords->width;
+	GanvEdge*        edge   = GANV_EDGE(item);
+	GanvEdgePrivate* impl   = edge->impl;
+	GanvEdgeCoords*  coords = &impl->coords;
+	const double     w      = coords->width;
 
 	if (coords->curved) {
 		*x1 = MIN(coords->x1, MIN(coords->cx1, MIN(coords->x2, coords->cx2))) - w;
@@ -252,7 +271,7 @@ ganv_edge_bounds(GanvItem* item,
 void
 ganv_edge_get_coords(const GanvEdge* edge, GanvEdgeCoords* coords)
 {
-	GanvEdgeImpl* impl = edge->impl;
+	GanvEdgePrivate* impl = edge->impl;
 
 	GANV_NODE_GET_CLASS(impl->tail)->tail_vector(
 		impl->tail, impl->head,
@@ -279,12 +298,8 @@ ganv_edge_get_coords(const GanvEdge* edge, GanvEdgeCoords* coords)
 static void
 ganv_edge_update(GanvItem* item, int flags)
 {
-	GanvEdge*     edge = GANV_EDGE(item);
-	GanvEdgeImpl* impl = edge->impl;
-
-	if (parent_class->update) {
-		(*parent_class->update)(item, flags);
-	}
+	GanvEdge*        edge = GANV_EDGE(item);
+	GanvEdgePrivate* impl = edge->impl;
 
 	// Request redraw of old location
 	ganv_edge_request_redraw(item, &impl->old_coords);
@@ -316,14 +331,16 @@ ganv_edge_update(GanvItem* item, int flags)
 
 	// Request redraw of new location
 	ganv_edge_request_redraw(item, &impl->coords);
+
+	parent_class->update(item, flags);
 }
 
 static void
 ganv_edge_draw(GanvItem* item,
                cairo_t* cr, double cx, double cy, double cw, double ch)
 {
-	GanvEdge*     edge = GANV_EDGE(item);
-	GanvEdgeImpl* impl = edge->impl;
+	GanvEdge*        edge = GANV_EDGE(item);
+	GanvEdgePrivate* impl = edge->impl;
 
 	double src_x = impl->coords.x1;
 	double src_y = impl->coords.y1;
@@ -334,7 +351,7 @@ ganv_edge_draw(GanvItem* item,
 
 	double r, g, b, a;
 	if (impl->highlighted) {
-		color_to_rgba(highlight_color(impl->color, 0x20), &r, &g, &b, &a);
+		color_to_rgba(highlight_color(impl->color, 0x40), &r, &g, &b, &a);
 	} else {
 		color_to_rgba(impl->color, &r, &g, &b, &a);
 	}
@@ -377,12 +394,6 @@ ganv_edge_draw(GanvItem* item,
 		cairo_curve_to(cr, src_x1, src_y1, src_x2, src_y2, join_x, join_y);
 		cairo_curve_to(cr, dst_x2, dst_y2, dst_x1, dst_y1, dst_x, dst_y);
 
-		if (impl->coords.arrowhead) {
-			cairo_line_to(cr, dst_x - 12, dst_y - 4);
-			cairo_move_to(cr, dst_x, dst_y);
-			cairo_line_to(cr, dst_x - 12, dst_y + 4);
-		}
-
 #ifdef GANV_DEBUG_CURVES
 		cairo_stroke(cr);
 		cairo_save(cr);
@@ -415,29 +426,44 @@ ganv_edge_draw(GanvItem* item,
 		cairo_restore(cr);
 #endif
 
+		cairo_stroke(cr);
+		if (impl->coords.arrowhead) {
+			cairo_move_to(cr, dst_x - 12, dst_y - 4);
+			cairo_line_to(cr, dst_x, dst_y);
+			cairo_line_to(cr, dst_x - 12, dst_y + 4);
+			cairo_close_path(cr);
+			cairo_stroke_preserve(cr);
+			cairo_fill(cr);
+		}
+
 	} else {
 		// Straight line from (x1, y1) to (x2, y2)
 		cairo_move_to(cr, src_x, src_y);
 		cairo_line_to(cr, dst_x, dst_y);
+		cairo_stroke(cr);
 
 		if (impl->coords.arrowhead) {
 			const double ah  = sqrt(dx * dx + dy * dy);
-			const double adx = dx / ah * 10.0;
-			const double ady = dy / ah * 10.0;
+			const double adx = dx / ah * 8.0;
+			const double ady = dy / ah * 8.0;
 
-			cairo_line_to(cr,
+			cairo_move_to(cr,
 			              dst_x + adx - ady/1.5,
 			              dst_y + ady + adx/1.5);
-			cairo_move_to(cr, dst_x, dst_y);
+			cairo_set_line_join(cr, CAIRO_LINE_JOIN_BEVEL);
+			cairo_line_to(cr, dst_x, dst_y);
+			cairo_set_line_join(cr, CAIRO_LINE_JOIN_MITER);
 			cairo_line_to(cr,
 			              dst_x + adx + ady/1.5,
 			              dst_y + ady - adx/1.5);
+			cairo_close_path(cr);
+			cairo_stroke_preserve(cr);
+			cairo_fill(cr);
 		}
 	}
 
-	cairo_stroke(cr);
-
-	if (impl->coords.handle_radius > 0.0) {
+	if (!ganv_canvas_exporting(item->impl->canvas) &&
+	    impl->coords.handle_radius > 0.0) {
 		cairo_move_to(cr, join_x, join_y);
 		cairo_arc(cr, join_x, join_y, impl->coords.handle_radius, 0, 2 * G_PI);
 		cairo_fill(cr);
@@ -490,8 +516,6 @@ ganv_edge_class_init(GanvEdgeClass* klass)
 
 	parent_class = GANV_ITEM_CLASS(g_type_class_peek_parent(klass));
 
-	g_type_class_add_private(klass, sizeof(GanvEdgeImpl));
-
 	gobject_class->set_property = ganv_edge_set_property;
 	gobject_class->get_property = ganv_edge_get_property;
 
@@ -553,7 +577,15 @@ ganv_edge_class_init(GanvEdgeClass* klass)
 			_("Color"),
 			_("Line color as an RGBA integer."),
 			0, G_MAXUINT,
-			0xA0A0A0FF,
+			0,
+			G_PARAM_READWRITE));
+
+	g_object_class_install_property(
+		gobject_class, PROP_CONSTRAINING, g_param_spec_boolean(
+			"constraining",
+			_("Constraining"),
+			_("Whether edge should constrain the layout."),
+			1,
 			G_PARAM_READWRITE));
 
 	g_object_class_install_property(
@@ -626,6 +658,13 @@ ganv_edge_new(GanvCanvas* canvas,
 	edge->impl->tail = tail;
 	edge->impl->head = head;
 
+	if (!edge->impl->color) {
+		const guint tail_color = GANV_NODE(tail)->impl->fill_color;
+		g_object_set(G_OBJECT(edge),
+		             "color", EDGE_COLOR(tail_color),
+		             NULL);
+	}
+
 	if (!edge->impl->ghost) {
 		ganv_canvas_add_edge(canvas, edge);
 	}
@@ -638,6 +677,19 @@ ganv_edge_update_location(GanvEdge* edge)
 	ganv_item_request_update(GANV_ITEM(edge));
 }
 
+gboolean
+ganv_edge_get_constraining(const GanvEdge* edge)
+{
+	return edge->impl->coords.constraining;
+}
+
+void
+ganv_edge_set_constraining(GanvEdge* edge, gboolean constraining)
+{
+	edge->impl->coords.constraining = constraining;
+	ganv_edge_request_redraw(GANV_ITEM(edge), &edge->impl->coords);
+}
+
 gboolean
 ganv_edge_get_curved(const GanvEdge* edge)
 {
@@ -674,6 +726,18 @@ ganv_edge_unselect(GanvEdge* edge)
 	ganv_edge_set_selected(edge, FALSE);
 }
 
+void
+ganv_edge_highlight(GanvEdge* edge)
+{
+	ganv_edge_set_highlighted(edge, TRUE);
+}
+
+void
+ganv_edge_unhighlight(GanvEdge* edge)
+{
+	ganv_edge_set_highlighted(edge, FALSE);
+}
+
 void
 ganv_edge_set_highlighted(GanvEdge* edge, gboolean highlighted)
 {
