@@ -1,5 +1,5 @@
 /* This file is part of Ganv.
- * Copyright 2007-2013 David Robillard <http://drobilla.net>
+ * Copyright 2007-2016 David Robillard <http://drobilla.net>
  *
  * Ganv is free software: you can redistribute it and/or modify it under the
  * terms of the GNU General Public License as published by the Free Software
@@ -42,7 +42,7 @@
 enum {
 	GANV_CANVAS_UPDATE_REQUESTED  = 1 << 0,
 	GANV_CANVAS_UPDATE_AFFINE     = 1 << 1,
-	GANV_CANVAS_UPDATE_VISIBILITY = 1 << 2,
+	GANV_CANVAS_UPDATE_VISIBILITY = 1 << 2
 };
 
 #define GCI_UPDATE_MASK (GANV_CANVAS_UPDATE_REQUESTED \
@@ -64,7 +64,8 @@ enum {
 
 static guint item_signals[ITEM_LAST_SIGNAL];
 
-G_DEFINE_TYPE(GanvItem, ganv_item, GTK_TYPE_OBJECT)
+G_DEFINE_TYPE_WITH_CODE(GanvItem, ganv_item, GTK_TYPE_OBJECT,
+                        G_ADD_PRIVATE(GanvItem))
 
 static GtkObjectClass* item_parent_class;
 
@@ -72,12 +73,12 @@ static GtkObjectClass* item_parent_class;
 static void
 ganv_item_init(GanvItem* item)
 {
-	GanvItemImpl* impl = G_TYPE_INSTANCE_GET_PRIVATE(
-		item, GANV_TYPE_ITEM, GanvItemImpl);
+	GanvItemPrivate* impl = ganv_item_get_instance_private(item);
 
 	item->object.flags |= GANV_ITEM_VISIBLE;
 	item->impl          = impl;
 	item->impl->managed = FALSE;
+	item->impl->wrapper = NULL;
 }
 
 /**
@@ -218,9 +219,10 @@ ganv_item_construct(GanvItem* item, GanvItem* parent,
 {
 	g_return_if_fail(GANV_IS_ITEM(item));
 
-	item->impl->parent = parent;
-	item->impl->canvas = item->impl->parent->impl->canvas;
-	item->impl->layer = 0;
+	item->impl->parent  = parent;
+	item->impl->wrapper = NULL;
+	item->impl->canvas  = item->impl->parent->impl->canvas;
+	item->impl->layer   = 0;
 
 	g_object_set_valist(G_OBJECT(item), first_arg_name, args);
 
@@ -343,6 +345,7 @@ ganv_item_invoke_update(GanvItem* item, int flags)
 	if (child_flags & GCI_UPDATE_MASK) {
 		if (GANV_ITEM_GET_CLASS(item)->update) {
 			GANV_ITEM_GET_CLASS(item)->update(item, child_flags);
+			g_assert(!(GTK_OBJECT_FLAGS(item) & GANV_ITEM_NEED_UPDATE));
 		}
 	}
 }
@@ -595,12 +598,6 @@ ganv_item_get_bounds(GanvItem* item, double* x1, double* y1, double* x2, double*
 void
 ganv_item_request_update(GanvItem* item)
 {
-	/* Note: At some point, if this short-circuit was enabled, the canvas stopped
-	   updating items entirely when connected modules are removed. */
-	if (item->object.flags & GANV_ITEM_NEED_UPDATE) {
-		return;
-	}
-
 	if (!item->impl->canvas) {
 		/* Item is being / has been destroyed, ignore */
 		return;
@@ -608,7 +605,8 @@ ganv_item_request_update(GanvItem* item)
 
 	item->object.flags |= GANV_ITEM_NEED_UPDATE;
 
-	if (item->impl->parent != NULL) {
+	if (item->impl->parent != NULL &&
+	    !(item->impl->parent->object.flags & GANV_ITEM_NEED_UPDATE)) {
 		/* Recurse up the tree */
 		ganv_item_request_update(item->impl->parent);
 	} else {
@@ -617,6 +615,18 @@ ganv_item_request_update(GanvItem* item)
 	}
 }
 
+void
+ganv_item_set_wrapper(GanvItem* item, void* wrapper)
+{
+	item->impl->wrapper = wrapper;
+}
+
+void*
+ganv_item_get_wrapper(GanvItem* item)
+{
+	return item->impl->wrapper;
+}
+
 static gboolean
 boolean_handled_accumulator(GSignalInvocationHint* ihint,
                             GValue*                return_accu,
@@ -643,8 +653,6 @@ ganv_item_class_init(GanvItemClass* klass)
 
 	item_parent_class = (GtkObjectClass*)g_type_class_peek_parent(klass);
 
-	g_type_class_add_private(klass, sizeof(GanvItemImpl));
-
 	gobject_class->set_property = ganv_item_set_property;
 	gobject_class->get_property = ganv_item_get_property;
 
@@ -652,7 +660,7 @@ ganv_item_class_init(GanvItemClass* klass)
 		(gobject_class, ITEM_PROP_PARENT,
 		 g_param_spec_object("parent", NULL, NULL,
 		                     GANV_TYPE_ITEM,
-		                     (G_PARAM_READABLE | G_PARAM_WRITABLE)));
+		                     (GParamFlags)(G_PARAM_READABLE | G_PARAM_WRITABLE)));
 
 	g_object_class_install_property
 		(gobject_class, ITEM_PROP_X,
@@ -660,14 +668,14 @@ ganv_item_class_init(GanvItemClass* klass)
 		                     _("X"),
 		                     _("X"),
 		                     -G_MAXDOUBLE, G_MAXDOUBLE, 0.0,
-		                     (G_PARAM_READABLE | G_PARAM_WRITABLE)));
+		                     (GParamFlags)(G_PARAM_READABLE | G_PARAM_WRITABLE)));
 	g_object_class_install_property
 		(gobject_class, ITEM_PROP_Y,
 		 g_param_spec_double("y",
 		                     _("Y"),
 		                     _("Y"),
 		                     -G_MAXDOUBLE, G_MAXDOUBLE, 0.0,
-		                     (G_PARAM_READABLE | G_PARAM_WRITABLE)));
+		                     (GParamFlags)(G_PARAM_READABLE | G_PARAM_WRITABLE)));
 
 	g_object_class_install_property
 		(gobject_class, ITEM_PROP_MANAGED,
@@ -675,7 +683,7 @@ ganv_item_class_init(GanvItemClass* klass)
 		                      _("Managed"),
 		                      _("Whether the item is managed by its parent"),
 		                      0,
-		                      (G_PARAM_READABLE | G_PARAM_WRITABLE)));
+		                      (GParamFlags)(G_PARAM_READABLE | G_PARAM_WRITABLE)));
 
 	item_signals[ITEM_EVENT]
 		= g_signal_new("event",
