@@ -1,5 +1,5 @@
 /* This file is part of Ganv.
- * Copyright 2007-2013 David Robillard <http://drobilla.net>
+ * Copyright 2007-2016 David Robillard <http://drobilla.net>
  *
  * Ganv is free software: you can redistribute it and/or modify it under the
  * terms of the GNU General Public License as published by the Free Software
@@ -34,21 +34,23 @@
 	for (const GanvPort** i = (const GanvPort**)ports->pdata; \
 	     i != (const GanvPort**)ports->pdata + ports->len; ++i)
 
+static const double PAD              = 2.0;
+static const double EDGE_PAD         = 5.0;
 static const double MODULE_LABEL_PAD = 2.0;
 
-G_DEFINE_TYPE(GanvModule, ganv_module, GANV_TYPE_BOX)
+G_DEFINE_TYPE_WITH_CODE(GanvModule, ganv_module, GANV_TYPE_BOX,
+                        G_ADD_PRIVATE(GanvModule))
 
 static GanvBoxClass* parent_class;
 
 enum {
-	PROP_0,
+	PROP_0
 };
 
 static void
 ganv_module_init(GanvModule* module)
 {
-	GanvModuleImpl* impl = G_TYPE_INSTANCE_GET_PRIVATE(
-		module, GANV_TYPE_MODULE, GanvModuleImpl);
+	GanvModulePrivate* impl = ganv_module_get_instance_private(module);
 
 	module->impl = impl;
 
@@ -61,7 +63,7 @@ ganv_module_init(GanvModule* module)
 	impl->embed_height  = 0;
 	impl->widest_input  = 0.0;
 	impl->widest_output = 0.0;
-	impl->must_resize   = TRUE;
+	impl->must_reorder  = FALSE;
 }
 
 static void
@@ -70,8 +72,8 @@ ganv_module_destroy(GtkObject* object)
 	g_return_if_fail(object != NULL);
 	g_return_if_fail(GANV_IS_MODULE(object));
 
-	GanvModule*     module = GANV_MODULE(object);
-	GanvModuleImpl* impl   = module->impl;
+	GanvModule*        module = GANV_MODULE(object);
+	GanvModulePrivate* impl   = module->impl;
 
 	if (impl->ports) {
 		FOREACH_PORT(impl->ports, p) {
@@ -81,6 +83,11 @@ ganv_module_destroy(GtkObject* object)
 		impl->ports = NULL;
 	}
 
+	if (impl->embed_item) {
+		g_object_unref(GTK_OBJECT(impl->embed_item));
+		impl->embed_item = NULL;
+	}
+
 	if (GTK_OBJECT_CLASS(parent_class)->destroy) {
 		(*GTK_OBJECT_CLASS(parent_class)->destroy)(object);
 	}
@@ -148,23 +155,35 @@ measure(GanvModule* module, Metrics* m)
 	double title_w, title_h;
 	title_size(module, &title_w, &title_h);
 
-	GanvCanvas*     canvas       = ganv_item_get_canvas(GANV_ITEM(module));
-	GanvText*       canvas_title = GANV_NODE(module)->impl->label;
-	GanvModuleImpl* impl         = module->impl;
+	GanvCanvas*        canvas       = ganv_item_get_canvas(GANV_ITEM(module));
+	GanvText*          canvas_title = GANV_NODE(module)->impl->label;
+	GanvModulePrivate* impl         = module->impl;
 
 	if (ganv_canvas_get_direction(canvas) == GANV_DIRECTION_DOWN) {
-		static const double PAD = 2.0;
-
-		double contents_width = PAD;
+		double contents_width = 0.0;
 		if (canvas_title) {
-			contents_width += title_w;
+			contents_width += title_w + (2.0 * PAD);
 		}
 
 		m->embed_x      = 0;
 		m->input_width  = ganv_module_get_empty_port_breadth(module);
 		m->output_width = ganv_module_get_empty_port_breadth(module);
 
-		const double ports_width = PAD + ((m->input_width + PAD) * impl->ports->len);
+		// TODO: cache this or merge with resize_right
+		unsigned n_inputs  = 0;
+		unsigned n_outputs = 0;
+		FOREACH_PORT(impl->ports, pi) {
+			if ((*pi)->impl->is_input) {
+				++n_inputs;
+			} else {
+				++n_outputs;
+			}
+		}
+
+		const unsigned hor_ports = MAX(1, MAX(n_inputs, n_outputs));
+		const double ports_width = (2 * EDGE_PAD) +
+			((m->input_width) * hor_ports) +
+			((PAD + 1.0) * (hor_ports - 1));
 
 		m->width = MAX(contents_width, ports_width);
 		m->width = MAX(m->width, impl->embed_width);
@@ -182,9 +201,11 @@ measure(GanvModule* module, Metrics* m)
 
 	m->width = (canvas_title) ? title_w + 10.0 : 1.0;
 
-	// Title is wide, put inputs and outputs beside each other
-	m->horiz = (impl->widest_input + impl->widest_output + 10.0
-	            < MAX(m->width, impl->embed_width));
+	// Title is wide or there is an embedded widget,
+	// put inputs and outputs beside each other
+	m->horiz = (impl->embed_item ||
+	            (impl->widest_input + impl->widest_output + 10.0
+	             < MAX(m->width, impl->embed_width)));
 
 	// Fit ports to module (or vice-versa)
 	m->input_width  = impl->widest_input;
@@ -237,24 +258,23 @@ place_title(GanvModule* module, GanvDirection dir)
 
 	if (!canvas_title) {
 		return;
-	} else if (dir == GANV_DIRECTION_RIGHT) {
-		ganv_item_set(GANV_ITEM(canvas_title),
-		              "x", (ganv_box_get_width(box) - title_w) / 2.0,
-		              "y", 2.0,
-		              NULL);
+	}
+
+	GanvItem* t = GANV_ITEM(canvas_title);
+	if (dir == GANV_DIRECTION_RIGHT) {
+		t->impl->x = (ganv_box_get_width(box) - title_w) / 2.0;
+		t->impl->y = 1.0;
 	} else {
-		ganv_item_set(GANV_ITEM(canvas_title),
-		              "x", (ganv_box_get_width(box) - title_w) / 2.0,
-		              "y", ganv_module_get_empty_port_depth(module) + 1.0,
-		              NULL);
+		t->impl->x = (ganv_box_get_width(box) - title_w) / 2.0;
+		t->impl->y = ganv_module_get_empty_port_depth(module) + 1.0;
 	}
 }
 
 static void
 resize_right(GanvModule* module)
 {
-	GanvCanvas*     canvas = ganv_item_get_canvas(GANV_ITEM(module));
-	GanvModuleImpl* impl   = module->impl;
+	GanvCanvas*        canvas = ganv_item_get_canvas(GANV_ITEM(module));
+	GanvModulePrivate* impl   = module->impl;
 
 	Metrics m;
 	measure(module, &m);
@@ -263,11 +283,12 @@ resize_right(GanvModule* module)
 	title_size(module, &title_w, &title_h);
 
 	// Basic height contains title
-	double header_height = 2.0 + title_h;
+	double header_height = title_h ? (3.0 + title_h) : EDGE_PAD;
 
 	if (impl->embed_item) {
 		ganv_item_set(impl->embed_item,
 		              "x", (double)m.embed_x,
+		              "y", header_height,
 		              NULL);
 	}
 
@@ -280,46 +301,43 @@ resize_right(GanvModule* module)
 	}
 
 	// Move ports to appropriate locations
-	int      i              = 0;
-	gboolean last_was_input = FALSE;
-	double   y              = 0.0;
-	double   h              = 0.0;
+	double   in_y           = header_height;
+	double   out_y          = header_height;
 	FOREACH_PORT(impl->ports, pi) {
 		GanvPort* const p     = (*pi);
 		GanvBox*  const pbox  = GANV_BOX(p);
 		GanvNode* const pnode = GANV_NODE(p);
-		h = ganv_box_get_height(pbox);
+		const double    h     = ganv_box_get_height(pbox);
+
+		// Offset to shift ports to make borders line up
+		const double border_off = (GANV_NODE(module)->impl->border_width -
+		                           pnode->impl->border_width) / 2.0;
 
 		if (p->impl->is_input) {
-			y = header_height + 2.0 + (i * (h + 2.0));
-			++i;
-			ganv_node_move_to(pnode, 0.0, y);
+			ganv_node_move_to(pnode, -border_off, in_y + 1.0);
 			ganv_box_set_width(pbox, m.input_width);
-			last_was_input = TRUE;
+			in_y += h + pnode->impl->border_width + 1.0;
 
 			ganv_canvas_for_each_edge_to(
 				canvas, pnode,
 				(GanvEdgeFunc)ganv_edge_update_location, NULL);
 		} else {
-			if (!m.horiz || !last_was_input) {
-				y = header_height + 2.0 + (i * (h + 2.0));
-				++i;
-			}
-			ganv_node_move_to(pnode, m.width - m.output_width, y);
+			ganv_node_move_to(pnode, m.width - m.output_width + border_off, out_y + 1.0);
 			ganv_box_set_width(pbox, m.output_width);
-			last_was_input = FALSE;
+			out_y += h + pnode->impl->border_width + 1.0;
 
 			ganv_canvas_for_each_edge_from(
 				canvas, pnode,
 				(GanvEdgeFunc)ganv_edge_update_location, NULL);
 		}
+
+		if (!m.horiz) {
+			in_y  = MAX(in_y, out_y);
+			out_y = MAX(in_y, out_y);
+		}
 	}
 
-	if (impl->ports->len == 0) {
-		h += header_height;
-	}
-
-	double height = y + h + 4.0;
+	double height = MAX(in_y, out_y) + EDGE_PAD;
 	if (impl->embed_item && m.embed_between)
 		height = MAX(height, impl->embed_height + header_height + 2.0);
 
@@ -331,8 +349,8 @@ resize_right(GanvModule* module)
 static void
 resize_down(GanvModule* module)
 {
-	GanvCanvas*     canvas = ganv_item_get_canvas(GANV_ITEM(module));
-	GanvModuleImpl* impl   = module->impl;
+	GanvCanvas*        canvas = ganv_item_get_canvas(GANV_ITEM(module));
+	GanvModulePrivate* impl   = module->impl;
 
 	Metrics m;
 	measure(module, &m);
@@ -340,8 +358,6 @@ resize_down(GanvModule* module)
 	double title_w, title_h;
 	title_size(module, &title_w, &title_h);
 
-	static const double PAD = 2.0;
-
 	const double port_depth   = ganv_module_get_empty_port_depth(module);
 	const double port_breadth = ganv_module_get_empty_port_breadth(module);
 
@@ -366,15 +382,20 @@ resize_down(GanvModule* module)
 		GanvNode* const pnode = GANV_NODE(p);
 		ganv_box_set_width(pbox, port_breadth);
 		ganv_box_set_height(pbox, port_depth);
+
+		// Offset to shift ports to make borders line up
+		const double border_off = (GANV_NODE(module)->impl->border_width -
+		                           pnode->impl->border_width) / 2.0;
+
 		if (p->impl->is_input) {
-			in_x = PAD + (in_count++ * (port_breadth + PAD));
-			ganv_node_move_to(pnode, in_x, 0);
+			in_x = EDGE_PAD + (in_count++ * (port_breadth + PAD + 1.0));
+			ganv_node_move_to(pnode, in_x, -border_off);
 			ganv_canvas_for_each_edge_to(
 				canvas, pnode,
 				(GanvEdgeFunc)ganv_edge_update_location, NULL);
 		} else {
-			out_x = PAD + (out_count++ * (port_breadth + PAD));
-			ganv_node_move_to(pnode, out_x, height - ganv_box_get_height(pbox));
+			out_x = EDGE_PAD + (out_count++ * (port_breadth + PAD + 1.0));
+			ganv_node_move_to(pnode, out_x, height - port_depth + border_off);
 			ganv_canvas_for_each_edge_from(
 				canvas, pnode,
 				(GanvEdgeFunc)ganv_edge_update_location, NULL);
@@ -389,7 +410,7 @@ resize_down(GanvModule* module)
 static void
 measure_ports(GanvModule* module)
 {
-	GanvModuleImpl* impl = module->impl;
+	GanvModulePrivate* impl = module->impl;
 
 	impl->widest_input  = 0.0;
 	impl->widest_output = 0.0;
@@ -409,12 +430,11 @@ measure_ports(GanvModule* module)
 }
 
 static void
-layout(GanvNode* self)
+ganv_module_resize(GanvNode* self)
 {
-	GanvModule*     module = GANV_MODULE(self);
-	GanvModuleImpl* impl   = module->impl;
-	GanvNode*       node   = GANV_NODE(self);
-	GanvCanvas*     canvas = ganv_item_get_canvas(GANV_ITEM(module));
+	GanvModule* module = GANV_MODULE(self);
+	GanvNode*   node   = GANV_NODE(self);
+	GanvCanvas* canvas = ganv_item_get_canvas(GANV_ITEM(module));
 
 	double label_w = 0.0;
 	double label_h = 0.0;
@@ -439,16 +459,8 @@ layout(GanvNode* self)
 		break;
 	}
 
-	impl->must_resize = FALSE;
-}
-
-static void
-ganv_module_resize(GanvNode* self)
-{
-	layout(self);
-
-	if (parent_class->parent_class.resize) {
-		parent_class->parent_class.resize(self);
+	if (GANV_NODE_CLASS(parent_class)->resize) {
+		GANV_NODE_CLASS(parent_class)->resize(self);
 	}
 }
 
@@ -468,9 +480,9 @@ static void
 ganv_module_add_port(GanvModule* module,
                      GanvPort*   port)
 {
-	GanvCanvas*     canvas = ganv_item_get_canvas(GANV_ITEM(module));
-	GanvModuleImpl* impl   = module->impl;
+	GanvModulePrivate* impl = module->impl;
 
+	// Update widest input/output measurements if necessary
 	const double width = ganv_port_get_natural_width(port);
 	if (port->impl->is_input && width > impl->widest_input) {
 		impl->widest_input = width;
@@ -478,11 +490,12 @@ ganv_module_add_port(GanvModule* module,
 		impl->widest_output = width;
 	}
 
-	impl->must_resize = TRUE;
-
+	// Add to port array
 	g_ptr_array_add(impl->ports, port);
-	place_title(module, ganv_canvas_get_direction(canvas));
-	ganv_item_request_update(GANV_ITEM(module));
+
+	// Request update with resize and reorder
+	GANV_NODE(module)->impl->must_resize = TRUE;
+	impl->must_reorder                   = TRUE;
 }
 
 static void
@@ -513,8 +526,7 @@ ganv_module_remove_port(GanvModule* module,
 			}
 		}
 
-		module->impl->must_resize = TRUE;
-		ganv_item_request_update(GANV_ITEM(module));
+		GANV_NODE(module)->impl->must_resize = TRUE;
 	} else {
 		fprintf(stderr, "Failed to find port to remove\n");
 	}
@@ -525,8 +537,10 @@ ganv_module_add(GanvItem* item, GanvItem* child)
 {
 	if (GANV_IS_PORT(child)) {
 		ganv_module_add_port(GANV_MODULE(item), GANV_PORT(child));
-	} else {
-		fprintf(stderr, "warning: Non-port item added to module.\n");
+	}
+	ganv_item_request_update(item);
+	if (GANV_ITEM_CLASS(parent_class)->add) {
+		GANV_ITEM_CLASS(parent_class)->add(item, child);
 	}
 }
 
@@ -535,23 +549,41 @@ ganv_module_remove(GanvItem* item, GanvItem* child)
 {
 	if (GANV_IS_PORT(child)) {
 		ganv_module_remove_port(GANV_MODULE(item), GANV_PORT(child));
-	} else {
-		fprintf(stderr, "warning: Non-port item removed from module.\n");
 	}
+	ganv_item_request_update(item);
+	if (GANV_ITEM_CLASS(parent_class)->remove) {
+		GANV_ITEM_CLASS(parent_class)->remove(item, child);
+	}
+}
+
+static int
+ptr_sort(const GanvPort** a, const GanvPort** b, const PortOrderCtx* ctx)
+{
+	return ctx->port_cmp(*a, *b, ctx->data);
 }
 
 static void
 ganv_module_update(GanvItem* item, int flags)
 {
-	GanvNode*   node   = GANV_NODE(item);
 	GanvModule* module = GANV_MODULE(item);
+	GanvCanvas* canvas = ganv_item_get_canvas(item);
 
-	if (module->impl->must_resize) {
-		layout(node);
+	if (module->impl->must_reorder) {
+		// Sort ports array
+		PortOrderCtx ctx = ganv_canvas_get_port_order(canvas);
+		if (ctx.port_cmp) {
+			g_ptr_array_sort_with_data(module->impl->ports,
+			                           (GCompareDataFunc)ptr_sort,
+
+			                           &ctx);
+		}
+		module->impl->must_reorder = FALSE;
 	}
 
-	GanvItemClass* item_class = GANV_ITEM_CLASS(parent_class);
-	item_class->update(item, flags);
+	if (module->impl->embed_item) {
+		// Kick the embedded item to update position if we have moved
+		ganv_item_move(GANV_ITEM(module->impl->embed_item), 0.0, 0.0);
+	}
 
 	FOREACH_PORT(module->impl->ports, p) {
 		ganv_item_invoke_update(GANV_ITEM(*p), flags);
@@ -560,6 +592,8 @@ ganv_module_update(GanvItem* item, int flags)
 	if (module->impl->embed_item) {
 		ganv_item_invoke_update(GANV_ITEM(module->impl->embed_item), flags);
 	}
+
+	GANV_ITEM_CLASS(parent_class)->update(item, flags);
 }
 
 static void
@@ -663,8 +697,6 @@ ganv_module_class_init(GanvModuleClass* klass)
 
 	parent_class = GANV_BOX_CLASS(g_type_class_peek_parent(klass));
 
-	g_type_class_add_private(klass, sizeof(GanvModuleImpl));
-
 	gobject_class->set_property = ganv_module_set_property;
 	gobject_class->get_property = ganv_module_get_property;
 
@@ -729,16 +761,15 @@ on_embed_size_request(GtkWidget*      widget,
                       GtkRequisition* r,
                       void*           user_data)
 {
-	GanvModule*     module = GANV_MODULE(user_data);
-	GanvModuleImpl* impl   = module->impl;
+	GanvModule*        module = GANV_MODULE(user_data);
+	GanvModulePrivate* impl   = module->impl;
 	if (impl->embed_width == r->width && impl->embed_height == r->height) {
 		return;
 	}
 
-	impl->embed_width  = r->width;
-	impl->embed_height = r->height;
-
-	impl->must_resize = TRUE;
+	impl->embed_width                    = r->width;
+	impl->embed_height                   = r->height;
+	GANV_NODE(module)->impl->must_resize = TRUE;
 
 	GtkAllocation allocation;
 	allocation.width = r->width;
@@ -755,44 +786,47 @@ void
 ganv_module_embed(GanvModule* module,
                   GtkWidget*  widget)
 {
-	GanvModuleImpl* impl = module->impl;
+	GanvModulePrivate* impl = module->impl;
+	if (!widget && !impl->embed_item) {
+		return;
+	}
 
 	if (impl->embed_item) {
+		// Free existing embedded widget
 		gtk_object_destroy(GTK_OBJECT(impl->embed_item));
 		impl->embed_item = NULL;
 	}
 
 	if (!widget) {
-		impl->embed_width  = 0;
-		impl->embed_height = 0;
-		impl->must_resize  = TRUE;
+		// Removing an existing embedded widget
+		impl->embed_width                    = 0;
+		impl->embed_height                   = 0;
+		GANV_NODE(module)->impl->must_resize = TRUE;
+		ganv_item_request_update(GANV_ITEM(module));
 		return;
 	}
 
 	double title_w, title_h;
 	title_size(module, &title_w, &title_h);
 
-	const double y = 4.0 + title_h;
 	impl->embed_item = ganv_item_new(
 		GANV_ITEM(module),
 		ganv_widget_get_type(),
 		"x", 2.0,
-		"y", y,
+		"y", 4.0 + title_h,
 		"widget", widget,
 		NULL);
 
-	gtk_widget_show_all(widget);
-
 	GtkRequisition r;
+	gtk_widget_show_all(widget);
 	gtk_widget_size_request(widget, &r);
 	on_embed_size_request(widget, &r, module);
-
 	ganv_item_show(impl->embed_item);
 
 	g_signal_connect(widget, "size-request",
 	                 G_CALLBACK(on_embed_size_request), module);
 
-	layout(GANV_NODE(module));
+	GANV_NODE(module)->impl->must_resize = TRUE;
 	ganv_item_request_update(GANV_ITEM(module));
 }
 
@@ -803,8 +837,8 @@ ganv_module_set_direction(GanvModule*   module,
 	FOREACH_PORT(module->impl->ports, p) {
 		ganv_port_set_direction(*p, direction);
 	}
-	module->impl->must_resize = TRUE;
-	ganv_module_resize(GANV_NODE(module));
+	GANV_NODE(module)->impl->must_resize = TRUE;
+	ganv_item_request_update(GANV_ITEM(module));
 }
 
 void
@@ -812,9 +846,9 @@ ganv_module_for_each_port(GanvModule*  module,
                           GanvPortFunc f,
                           void*        data)
 {
-	GanvModuleImpl* impl = module->impl;
-	const int       len  = impl->ports->len;
-	GanvPort**      copy = (GanvPort**)malloc(sizeof(GanvPort*) * len);
+	GanvModulePrivate* impl = module->impl;
+	const int          len  = impl->ports->len;
+	GanvPort**         copy = (GanvPort**)malloc(sizeof(GanvPort*) * len);
 	memcpy(copy, impl->ports->pdata, sizeof(GanvPort*) * len);
 
 	for (int i = 0; i < len; ++i) {
