@@ -1,5 +1,5 @@
 /* This file is part of Ganv.
- * Copyright 2007-2013 David Robillard <http://drobilla.net>
+ * Copyright 2007-2016 David Robillard <http://drobilla.net>
  *
  * Ganv is free software: you can redistribute it and/or modify it under the
  * terms of the GNU General Public License as published by the Free Software
@@ -24,7 +24,8 @@
 
 guint signal_moved;
 
-G_DEFINE_TYPE(GanvNode, ganv_node, GANV_TYPE_ITEM)
+G_DEFINE_TYPE_WITH_CODE(GanvNode, ganv_node, GANV_TYPE_ITEM,
+                        G_ADD_PRIVATE(GanvNode))
 
 static GanvItemClass* parent_class;
 
@@ -51,8 +52,7 @@ enum {
 static void
 ganv_node_init(GanvNode* node)
 {
-	GanvNodeImpl* impl = G_TYPE_INSTANCE_GET_PRIVATE(
-		node, GANV_TYPE_NODE, GanvNodeImpl);
+	GanvNodePrivate* impl = ganv_node_get_instance_private(node);
 
 	node->impl = impl;
 
@@ -71,6 +71,7 @@ ganv_node_init(GanvNode* node)
 	impl->draggable    = FALSE;
 	impl->show_label   = TRUE;
 	impl->grabbed      = FALSE;
+	impl->must_resize  = FALSE;
 #ifdef GANV_FDGL
 	impl->force.x       = 0.0;
 	impl->force.y       = 0.0;
@@ -93,8 +94,8 @@ ganv_node_destroy(GtkObject* object)
 	g_return_if_fail(object != NULL);
 	g_return_if_fail(GANV_IS_NODE(object));
 
-	GanvNode*     node = GANV_NODE(object);
-	GanvNodeImpl* impl = node->impl;
+	GanvNode*        node = GANV_NODE(object);
+	GanvNodePrivate* impl = node->impl;
 	if (impl->label) {
 		g_object_unref(impl->label);
 		impl->label = NULL;
@@ -114,6 +115,22 @@ ganv_node_destroy(GtkObject* object)
 	item->impl->canvas = NULL;
 }
 
+static void
+ganv_node_update(GanvItem* item, int flags)
+{
+	GanvNode* node = GANV_NODE(item);
+	if (node->impl->must_resize) {
+		ganv_node_resize(node);
+		node->impl->must_resize = FALSE;
+	}
+
+	if (node->impl->label) {
+		ganv_item_invoke_update(GANV_ITEM(node->impl->label), flags);
+	}
+
+	GANV_ITEM_CLASS(parent_class)->update(item, flags);
+}
+
 static void
 ganv_node_draw(GanvItem* item,
                cairo_t* cr, double cx, double cy, double cw, double ch)
@@ -133,11 +150,10 @@ ganv_node_set_property(GObject*      object,
 	g_return_if_fail(object != NULL);
 	g_return_if_fail(GANV_IS_NODE(object));
 
-	GanvNode*     node = GANV_NODE(object);
-	GanvNodeImpl* impl = node->impl;
+	GanvNode*        node = GANV_NODE(object);
+	GanvNodePrivate* impl = node->impl;
 
 	switch (prop_id) {
-		SET_CASE(PARTNER, object, impl->partner);
 		SET_CASE(DASH_LENGTH, double, impl->dash_length);
 		SET_CASE(DASH_OFFSET, double, impl->dash_offset);
 		SET_CASE(BORDER_WIDTH, double, impl->border_width);
@@ -149,6 +165,9 @@ ganv_node_set_property(GObject*      object,
 		SET_CASE(HIGHLIGHTED, boolean, impl->highlighted);
 		SET_CASE(DRAGGABLE, boolean, impl->draggable);
 		SET_CASE(GRABBED, boolean, impl->grabbed);
+	case PROP_PARTNER:
+		impl->partner = (GanvNode*)g_value_get_object(value);
+		break;
 	case PROP_SELECTED:
 		if (impl->selected != g_value_get_boolean(value)) {
 			GanvItem* item = GANV_ITEM(object);
@@ -193,12 +212,12 @@ ganv_node_get_property(GObject*    object,
 	g_return_if_fail(object != NULL);
 	g_return_if_fail(GANV_IS_NODE(object));
 
-	GanvNode*     node = GANV_NODE(object);
-	GanvNodeImpl* impl = node->impl;
+	GanvNode*        node = GANV_NODE(object);
+	GanvNodePrivate* impl = node->impl;
 
 	switch (prop_id) {
 		GET_CASE(PARTNER, object, impl->partner);
-		GET_CASE(LABEL, string, impl->label->impl->text);
+		GET_CASE(LABEL, string, impl->label ? impl->label->impl->text : NULL);
 		GET_CASE(DASH_LENGTH, double, impl->dash_length);
 		GET_CASE(DASH_OFFSET, double, impl->dash_offset);
 		GET_CASE(BORDER_WIDTH, double, impl->border_width);
@@ -280,7 +299,7 @@ ganv_node_get_draw_properties(const GanvNode* node,
                               double*         border_color,
                               double*         fill_color)
 {
-	GanvNodeImpl* impl = node->impl;
+	GanvNodePrivate* impl = node->impl;
 
 	*dash_length  = impl->dash_length;
 	*border_color = impl->border_color;
@@ -288,18 +307,19 @@ ganv_node_get_draw_properties(const GanvNode* node,
 
 	if (impl->selected) {
 		*dash_length  = 4.0;
-		*border_color = highlight_color(impl->border_color, 0x20);
+		*border_color = highlight_color(impl->border_color, 0x40);
 	}
 
 	if (impl->highlighted) {
-		*border_color = highlight_color(impl->border_color, 0x20);
+		*border_color = highlight_color(impl->border_color, 0x40);
+		*fill_color   = impl->fill_color;
 	}
 }
 
 void
 ganv_node_set_label(GanvNode* node, const char* str)
 {
-	GanvNodeImpl* impl = node->impl;
+	GanvNodePrivate* impl = node->impl;
 	if (!str || str[0] == '\0') {
 		if (impl->label) {
 			gtk_object_destroy(GTK_OBJECT(impl->label));
@@ -313,19 +333,13 @@ ganv_node_set_label(GanvNode* node, const char* str)
 		impl->label = GANV_TEXT(ganv_item_new(GANV_ITEM(node),
 		                                      ganv_text_get_type(),
 		                                      "text", str,
-		                                      "color", 0xFFFFFFFF,
+		                                      "color", DEFAULT_TEXT_COLOR,
 		                                      "managed", TRUE,
 		                                      NULL));
 	}
 
-	if (impl->show_label) {
-		GanvNodeClass* klass = GANV_NODE_GET_CLASS(node);
-		if (klass->resize) {
-			klass->resize(node);
-		}
-
-		ganv_item_request_update(GANV_ITEM(node));
-	}
+	impl->must_resize = TRUE;
+	ganv_item_request_update(GANV_ITEM(node));
 }
 
 void
@@ -370,6 +384,7 @@ ganv_node_default_move(GanvNode* node,
 	ganv_item_move(GANV_ITEM(node), dx, dy);
 	ganv_canvas_for_each_edge_on(
 		canvas, node, (GanvEdgeFunc)ganv_edge_update_location, NULL);
+	ganv_item_request_update(GANV_ITEM(node));
 }
 
 static void
@@ -377,11 +392,10 @@ ganv_node_default_move_to(GanvNode* node,
                           double    x,
                           double    y)
 {
-	GanvCanvas* canvas = ganv_item_get_canvas(GANV_ITEM(node));
-	ganv_item_set(GANV_ITEM(node),
-	              "x", x,
-	              "y", y,
-	              NULL);
+	GanvItem*   item   = GANV_ITEM(node);
+	GanvCanvas* canvas = ganv_item_get_canvas(item);
+	item->impl->x = x;
+	item->impl->y = y;
 	if (node->impl->can_tail) {
 		ganv_canvas_for_each_edge_from(
 			canvas, node, (GanvEdgeFunc)ganv_edge_update_location, NULL);
@@ -389,6 +403,7 @@ ganv_node_default_move_to(GanvNode* node,
 		ganv_canvas_for_each_edge_to(
 			canvas, node, (GanvEdgeFunc)ganv_edge_update_location, NULL);
 	}
+	ganv_item_request_update(GANV_ITEM(node));
 }
 
 static void
@@ -398,6 +413,7 @@ ganv_node_default_resize(GanvNode* node)
 	if (GANV_IS_NODE(item->impl->parent)) {
 		ganv_node_resize(GANV_NODE(item->impl->parent));
 	}
+	node->impl->must_resize = FALSE;
 }
 
 static void
@@ -405,7 +421,8 @@ ganv_node_default_redraw_text(GanvNode* node)
 {
 	if (node->impl->label) {
 		ganv_text_layout(node->impl->label);
-		ganv_node_resize(node);
+		node->impl->must_resize = TRUE;
+		ganv_item_request_update(GANV_ITEM(node));
 	}
 }
 
@@ -424,12 +441,14 @@ ganv_node_default_event(GanvItem* item,
 	switch (event->type) {
 	case GDK_ENTER_NOTIFY:
 		ganv_item_raise(GANV_ITEM(node));
-		ganv_item_set(GANV_ITEM(node), "highlighted", TRUE, NULL);
+		node->impl->highlighted = TRUE;
+		ganv_item_request_update(item);
 		return TRUE;
 
 	case GDK_LEAVE_NOTIFY:
 		ganv_item_lower(GANV_ITEM(node));
-		ganv_item_set(GANV_ITEM(node), "highlighted", FALSE, NULL);
+		node->impl->highlighted = FALSE;
+		ganv_item_request_update(item);
 		return TRUE;
 
 	case GDK_BUTTON_PRESS:
@@ -526,8 +545,6 @@ ganv_node_class_init(GanvNodeClass* klass)
 
 	parent_class = GANV_ITEM_CLASS(g_type_class_peek_parent(klass));
 
-	g_type_class_add_private(klass, sizeof(GanvNodeImpl));
-
 	gobject_class->set_property = ganv_node_set_property;
 	gobject_class->get_property = ganv_node_get_property;
 
@@ -684,6 +701,7 @@ ganv_node_class_init(GanvNodeClass* klass)
 
 	item_class->realize = ganv_node_realize;
 	item_class->event   = ganv_node_default_event;
+	item_class->update  = ganv_node_update;
 	item_class->draw    = ganv_node_draw;
 
 	klass->disconnect  = ganv_node_default_disconnect;
@@ -855,6 +873,7 @@ void
 ganv_node_resize(GanvNode* node)
 {
 	GANV_NODE_GET_CLASS(node)->resize(node);
+	node->impl->must_resize = FALSE;
 }
 
 void
