@@ -1,5 +1,5 @@
 /* This file is part of Ganv.
- * Copyright 2007-2013 David Robillard <http://drobilla.net>
+ * Copyright 2007-2015 David Robillard <http://drobilla.net>
  *
  * Ganv is free software: you can redistribute it and/or modify it under the
  * terms of the GNU General Public License as published by the Free Software
@@ -28,13 +28,11 @@
 static const double PORT_LABEL_HPAD = 4.0;
 static const double PORT_LABEL_VPAD = 1.0;
 
-static const guchar check_off[] = { 0xE2, 0x98, 0x90, 0 };
-static const guchar check_on[]  = { 0xE2, 0x98, 0x91, 0 };
-
 static void
 ganv_port_update_control_slider(GanvPort* port, float value, gboolean force);
 
-G_DEFINE_TYPE(GanvPort, ganv_port, GANV_TYPE_BOX)
+G_DEFINE_TYPE_WITH_CODE(GanvPort, ganv_port, GANV_TYPE_BOX,
+                        G_ADD_PRIVATE(GanvPort))
 
 static GanvBoxClass* parent_class;
 
@@ -54,9 +52,10 @@ static guint port_signals[PORT_LAST_SIGNAL];
 static void
 ganv_port_init(GanvPort* port)
 {
-	port->impl = G_TYPE_INSTANCE_GET_PRIVATE(
-		port, GANV_TYPE_PORT, GanvPortImpl);
+	port->impl = ganv_port_get_instance_private(port);
 
+	port->impl->control         = NULL;
+	port->impl->value_label     = NULL;
 	port->impl->is_input        = TRUE;
 	port->impl->is_controllable = FALSE;
 }
@@ -125,6 +124,24 @@ ganv_port_get_property(GObject*    object,
 	}
 }
 
+static void
+ganv_port_update(GanvItem* item, int flags)
+{
+	GanvPort*        port = GANV_PORT(item);
+	GanvPortPrivate* impl = port->impl;
+
+	if (impl->control) {
+		ganv_item_invoke_update(GANV_ITEM(impl->control->rect), flags);
+	}
+
+	if (impl->value_label) {
+		ganv_item_invoke_update(GANV_ITEM(port->impl->value_label), flags);
+	}
+
+	GanvItemClass* item_class = GANV_ITEM_CLASS(parent_class);
+	item_class->update(item, flags);
+}
+
 static void
 ganv_port_draw(GanvItem* item,
                cairo_t* cr, double cx, double cy, double cw, double ch)
@@ -137,8 +154,22 @@ ganv_port_draw(GanvItem* item,
 	item_class->draw(item, cr, cx, cy, cw, ch);
 
 	if (port->impl->control) {
+		// Clip to port boundaries (to stay within radiused borders)
+		cairo_save(cr);
+		const double  pad    = GANV_NODE(port)->impl->border_width / 2.0;
+		GanvBoxCoords coords = GANV_BOX(port)->impl->coords;
+		ganv_item_i2w_pair(GANV_ITEM(port),
+		                   &coords.x1, &coords.y1, &coords.x2, &coords.y2);
+		ganv_box_path(GANV_BOX(port), cr,
+		              coords.x1 + pad, coords.y1 + pad,
+		              coords.x2 - pad, coords.y2 - pad,
+		              -pad);
+		cairo_clip(cr);
+
 		GanvItem* const rect = GANV_ITEM(port->impl->control->rect);
 		GANV_ITEM_GET_CLASS(rect)->draw(rect, cr, cx, cy, cw, ch);
+
+		cairo_restore(cr);
 	}
 
 	if (ganv_canvas_get_direction(canvas) == GANV_DIRECTION_DOWN ||
@@ -148,7 +179,7 @@ ganv_port_draw(GanvItem* item,
 
 	GanvItem* labels[2] = {
 		GANV_ITEM(GANV_NODE(item)->impl->label),
-		port->impl->control ? GANV_ITEM(port->impl->control->label) : NULL
+		port->impl->value_label ? GANV_ITEM(port->impl->value_label) : NULL
 	};
 	for (int i = 0; i < 2; ++i) {
 		if (labels[i] && (labels[i]->object.flags & GANV_ITEM_VISIBLE)) {
@@ -170,19 +201,20 @@ ganv_port_tail_vector(const GanvNode* self,
 	GanvItem*   item   = &port->box.node.item;
 	GanvCanvas* canvas = ganv_item_get_canvas(item);
 
-	const double px = item->impl->x;
-	const double py = item->impl->y;
+	const double px           = item->impl->x;
+	const double py           = item->impl->y;
+	const double border_width = GANV_NODE(port)->impl->border_width;
 
 	switch (ganv_canvas_get_direction(canvas)) {
 	case GANV_DIRECTION_RIGHT:
-		*x  = px + ganv_box_get_width(&port->box);
+		*x  = px + ganv_box_get_width(&port->box) + (border_width / 2.0);
 		*y  = py + ganv_box_get_height(&port->box) / 2.0;
 		*dx = 1.0;
 		*dy = 0.0;
 		break;
 	case GANV_DIRECTION_DOWN:
 		*x  = px + ganv_box_get_width(&port->box) / 2.0;
-		*y  = py + ganv_box_get_height(&port->box);
+		*y  = py + ganv_box_get_height(&port->box) + (border_width / 2.0);
 		*dx = 0.0;
 		*dy = 1.0;
 		break;
@@ -203,19 +235,20 @@ ganv_port_head_vector(const GanvNode* self,
 	GanvItem*   item   = &port->box.node.item;
 	GanvCanvas* canvas = ganv_item_get_canvas(item);
 
-	const double px = item->impl->x;
-	const double py = item->impl->y;
+	const double px           = item->impl->x;
+	const double py           = item->impl->y;
+	const double border_width = GANV_NODE(port)->impl->border_width;
 
 	switch (ganv_canvas_get_direction(canvas)) {
 	case GANV_DIRECTION_RIGHT:
-		*x  = px;
+		*x  = px - (border_width / 2.0);
 		*y  = py + ganv_box_get_height(&port->box) / 2.0;
 		*dx = -1.0;
 		*dy = 0.0;
 		break;
 	case GANV_DIRECTION_DOWN:
 		*x  = px + ganv_box_get_width(&port->box) / 2.0;
-		*y  = 0.0;
+		*y  = py - (border_width / 2.0);
 		*dx = 0.0;
 		*dy = -1.0;
 		break;
@@ -225,24 +258,36 @@ ganv_port_head_vector(const GanvNode* self,
 }
 
 static void
-ganv_port_place_value_label(GanvPort* port)
+ganv_port_place_labels(GanvPort* port)
 {
-	GanvPortControl* control = port->impl->control;
-	if (control && control->label) {
-		GanvCanvas*  canvas  = ganv_item_get_canvas(GANV_ITEM(port));
-		const double port_w  = ganv_box_get_width(&port->box);
-		const double label_w = control->label->impl->coords.width;
+	GanvCanvas*      canvas   = ganv_item_get_canvas(GANV_ITEM(port));
+	GanvPortPrivate* impl     = port->impl;
+	GanvText*        label    = GANV_NODE(port)->impl->label;
+	const double     port_w   = ganv_box_get_width(&port->box);
+	const double     port_h   = ganv_box_get_height(&port->box);
+	double           vlabel_w = 0.0;
+	if (impl->value_label) {
+		const double vlabel_h = impl->value_label->impl->coords.height;
+		vlabel_w = impl->value_label->impl->coords.width;
 		if (ganv_canvas_get_direction(canvas) == GANV_DIRECTION_RIGHT) {
-			ganv_item_set(GANV_ITEM(control->label),
-			              "x", port_w - label_w - PORT_LABEL_HPAD,
-			              "y", PORT_LABEL_VPAD,
+			ganv_item_set(GANV_ITEM(impl->value_label),
+			              "x", PORT_LABEL_HPAD,
+			              "y", (port_h - vlabel_h) / 2.0 - PORT_LABEL_VPAD,
 			              NULL);
 		} else {
-			const double port_h  = ganv_box_get_height(&port->box);
-			const double label_h = control->label->impl->coords.height;
-			ganv_item_set(GANV_ITEM(control->label),
-			              "x", (port_w - label_w) / 2.0,
-			              "y", (port_h - label_h) / 2.0,
+			ganv_item_set(GANV_ITEM(impl->value_label),
+			              "x", (port_w - vlabel_w) / 2.0,
+			              "y", (port_h - vlabel_h) / 2.0 - PORT_LABEL_VPAD,
+			              NULL);
+		}
+		vlabel_w += PORT_LABEL_HPAD;
+	}
+	if (label) {
+		const double label_h = label->impl->coords.height;
+		if (ganv_canvas_get_direction(canvas) == GANV_DIRECTION_RIGHT) {
+			ganv_item_set(GANV_ITEM(label),
+			              "x", vlabel_w + PORT_LABEL_HPAD,
+			              "y", (port_h - label_h) / 2.0 - PORT_LABEL_VPAD,
 			              NULL);
 		}
 	}
@@ -254,7 +299,7 @@ ganv_port_resize(GanvNode* self)
 	GanvPort* port   = GANV_PORT(self);
 	GanvNode* node   = GANV_NODE(self);
 	GanvText* label  = node->impl->label;
-	GanvText* vlabel = port->impl->control ? port->impl->control->label : NULL;
+	GanvText* vlabel = port->impl->value_label;
 
 	double label_w  = 0.0;
 	double label_h  = 0.0;
@@ -268,18 +313,19 @@ ganv_port_resize(GanvNode* self)
 	}
 
 	if (label || vlabel) {
-		ganv_box_set_width(&port->box,
-		                   label_w + vlabel_w + (PORT_LABEL_HPAD * 2.0));
+		double labels_w = label_w + PORT_LABEL_HPAD * 2.0;
+		if (vlabel_w != 0.0) {
+			labels_w += vlabel_w + PORT_LABEL_HPAD;
+		}
+		ganv_box_set_width(&port->box, labels_w);
 		ganv_box_set_height(&port->box,
 		                    MAX(label_h, vlabel_h) + (PORT_LABEL_VPAD * 2.0));
-		ganv_item_set(GANV_ITEM(node->impl->label),
-		              "x", PORT_LABEL_HPAD,
-		              "y", PORT_LABEL_VPAD,
-		              NULL);
+
+		ganv_port_place_labels(port);
 	}
 
-	if (parent_class->parent_class.resize) {
-		parent_class->parent_class.resize(self);
+	if (GANV_NODE_CLASS(parent_class)->resize) {
+		GANV_NODE_CLASS(parent_class)->resize(self);
 	}
 }
 
@@ -287,13 +333,13 @@ static void
 ganv_port_redraw_text(GanvNode* node)
 {
 	GanvPort* port = GANV_PORT(node);
-	if (port->impl->control && port->impl->control->label) {
-		ganv_text_layout(port->impl->control->label);
-		ganv_port_place_value_label(port);
+	if (port->impl->value_label) {
+		ganv_text_layout(port->impl->value_label);
 	}
 	if (GANV_NODE_CLASS(parent_class)->redraw_text) {
 		(*GANV_NODE_CLASS(parent_class)->redraw_text)(node);
 	}
+	ganv_port_place_labels(port);
 }
 
 static void
@@ -304,8 +350,8 @@ ganv_port_set_width(GanvBox* box,
 	parent_class->set_width(box, width);
 	if (port->impl->control) {
 		ganv_port_update_control_slider(port, port->impl->control->value, TRUE);
-		ganv_port_place_value_label(port);
 	}
+	ganv_port_place_labels(port);
 }
 
 static void
@@ -315,13 +361,12 @@ ganv_port_set_height(GanvBox* box,
 	GanvPort* port = GANV_PORT(box);
 	parent_class->set_height(box, height);
 	if (port->impl->control) {
-		double control_y1;
-		g_object_get(port->impl->control->rect, "y1", &control_y1, NULL);
 		ganv_item_set(GANV_ITEM(port->impl->control->rect),
-		              "y2", control_y1 + height,
+		              "y1", box->impl->coords.border_width / 2.0,
+		              "y2", height - box->impl->coords.border_width / 2.0,
 		              NULL);
-		ganv_port_place_value_label(port);
 	}
+	ganv_port_place_labels(port);
 }
 
 static gboolean
@@ -343,8 +388,6 @@ ganv_port_class_init(GanvPortClass* klass)
 
 	parent_class = GANV_BOX_CLASS(g_type_class_peek_parent(klass));
 
-	g_type_class_add_private(klass, sizeof(GanvPortImpl));
-
 	gobject_class->set_property = ganv_port_set_property;
 	gobject_class->get_property = ganv_port_get_property;
 
@@ -376,8 +419,9 @@ ganv_port_class_init(GanvPortClass* klass)
 
 	object_class->destroy = ganv_port_destroy;
 
-	item_class->event = ganv_port_event;
-	item_class->draw  = ganv_port_draw;
+	item_class->update = ganv_port_update;
+	item_class->event  = ganv_port_event;
+	item_class->draw   = ganv_port_draw;
 
 	node_class->tail_vector = ganv_port_tail_vector;
 	node_class->head_vector = ganv_port_head_vector;
@@ -412,7 +456,7 @@ ganv_port_new(GanvModule* module,
 	node->impl->can_tail     = !is_input;
 	node->impl->can_head     = is_input;
 	node->impl->draggable    = FALSE;
-	node->impl->border_width = 1.0;
+	node->impl->border_width = 2.0;
 
 	GanvCanvas* canvas = ganv_item_get_canvas(GANV_ITEM(port));
 	ganv_port_set_direction(port, ganv_canvas_get_direction(canvas));
@@ -429,44 +473,50 @@ ganv_port_set_direction(GanvPort*     port,
 	gboolean  is_input = port->impl->is_input;
 	switch (direction) {
 	case GANV_DIRECTION_RIGHT:
-		box->impl->radius_tl = (is_input ? 0.0 : 4.0);
-		box->impl->radius_tr = (is_input ? 4.0 : 0.0);
-		box->impl->radius_br = (is_input ? 4.0 : 0.0);
-		box->impl->radius_bl = (is_input ? 0.0 : 4.0);
+		box->impl->radius_tl = (is_input ? 0.0 : 5.0);
+		box->impl->radius_tr = (is_input ? 5.0 : 0.0);
+		box->impl->radius_br = (is_input ? 5.0 : 0.0);
+		box->impl->radius_bl = (is_input ? 0.0 : 5.0);
 		break;
 	case GANV_DIRECTION_DOWN:
-		box->impl->radius_tl = (is_input ? 0.0 : 4.0);
-		box->impl->radius_tr = (is_input ? 0.0 : 4.0);
-		box->impl->radius_br = (is_input ? 4.0 : 0.0);
-		box->impl->radius_bl = (is_input ? 4.0 : 0.0);
+		box->impl->radius_tl = (is_input ? 0.0 : 5.0);
+		box->impl->radius_tr = (is_input ? 0.0 : 5.0);
+		box->impl->radius_br = (is_input ? 5.0 : 0.0);
+		box->impl->radius_bl = (is_input ? 5.0 : 0.0);
 		break;
 	}
-	ganv_node_resize(node);
+
+	node->impl->must_resize = TRUE;
+	ganv_item_request_update(GANV_ITEM(node));
 }
 
 void
 ganv_port_show_control(GanvPort* port)
 {
+	if (port->impl->control) {
+		return;
+	}
+
+	const guint  color        = 0xFFFFFF66;
+	const double border_width = GANV_NODE(port)->impl->border_width;
+
 	GanvPortControl* control = (GanvPortControl*)malloc(sizeof(GanvPortControl));
 	port->impl->control = control;
 
-	guint control_col = highlight_color(GANV_NODE(port)->impl->fill_color, 0x40);
-
 	control->value      = 0.0f;
 	control->min        = 0.0f;
-	control->max        = 0.0f;
+	control->max        = 1.0f;
 	control->is_toggle  = FALSE;
 	control->is_integer = FALSE;
-	control->label      = NULL;
 	control->rect       = GANV_BOX(
 		ganv_item_new(GANV_ITEM(port),
 		              ganv_box_get_type(),
-		              "x1", 0.0,
-		              "y1", 0.0,
+		              "x1", border_width / 2.0,
+		              "y1", border_width / 2.0,
 		              "x2", 0.0,
-		              "y2", ganv_box_get_height(&port->box),
-		              "fill-color", control_col,
-		              "border-color", control_col,
+		              "y2", ganv_box_get_height(&port->box) - border_width / 2.0,
+		              "fill-color", color,
+		              "border-color", color,
 		              "border-width", 0.0,
 		              "managed", TRUE,
 		              NULL));
@@ -485,35 +535,31 @@ void
 ganv_port_set_value_label(GanvPort*   port,
                           const char* str)
 {
-	GanvPortImpl* impl = port->impl;
-	if (!impl->control) {
-		return;
-	}
+	GanvPortPrivate* impl = port->impl;
 
 	if (!str || str[0] == '\0') {
-		if (impl->control->label) {
-			gtk_object_destroy(GTK_OBJECT(impl->control->label));
-			impl->control->label = NULL;
+		if (impl->value_label) {
+			gtk_object_destroy(GTK_OBJECT(impl->value_label));
+			impl->value_label = NULL;
 		}
-	} else if (impl->control->label) {
-		ganv_item_set(GANV_ITEM(impl->control->label),
+	} else if (impl->value_label) {
+		ganv_item_set(GANV_ITEM(impl->value_label),
 		              "text", str,
 		              NULL);
 	} else {
-		impl->control->label = GANV_TEXT(ganv_item_new(GANV_ITEM(port),
-		                                               ganv_text_get_type(),
-		                                               "text", str,
-		                                               "color", 0xFFFFFFFF,
-		                                               "managed", TRUE,
-		                                               NULL));
-		ganv_port_resize(GANV_NODE(port));
+		impl->value_label = GANV_TEXT(ganv_item_new(GANV_ITEM(port),
+		                                            ganv_text_get_type(),
+		                                            "text", str,
+		                                            "color", DIM_TEXT_COLOR,
+		                                            "managed", TRUE,
+		                                            NULL));
 	}
 }
 
 static void
 ganv_port_update_control_slider(GanvPort* port, float value, gboolean force)
 {
-	GanvPortImpl* impl = port->impl;
+	GanvPortPrivate* impl = port->impl;
 	if (!impl->control) {
 		return;
 	}
@@ -536,14 +582,17 @@ ganv_port_update_control_slider(GanvPort* port, float value, gboolean force)
 	if (value > impl->control->max) {
 		value = impl->control->max;
 	}
-	
+
 	if (!force && value == impl->control->value) {
 		return;  // No change, do nothing
 	}
 
+	const double span = (ganv_box_get_width(&port->box) -
+	                     GANV_NODE(port)->impl->border_width);
+
 	const double w = (value - impl->control->min)
 		/ (impl->control->max - impl->control->min)
-		* ganv_box_get_width(&port->box);
+		* span;
 
 	if (isnan(w)) {
 		return;  // Shouldn't happen, but ignore crazy values
@@ -551,7 +600,7 @@ ganv_port_update_control_slider(GanvPort* port, float value, gboolean force)
 
 	// Redraw port
 	impl->control->value = value;
-	ganv_box_set_width(impl->control->rect, MAX(0.0, w - 1.0));
+	ganv_box_set_width(impl->control->rect, MAX(0.0, w));
 	ganv_box_request_redraw(
 		GANV_ITEM(port), &GANV_BOX(port)->impl->coords, FALSE);
 }
@@ -581,12 +630,6 @@ void
 ganv_port_set_control_value(GanvPort* port,
                             float     value)
 {
-	GanvPortImpl* impl = port->impl;
-
-	if (impl->control && impl->control->is_toggle) {
-		ganv_port_set_value_label(
-			port, (const char*)((value == 0.0f) ? check_off : check_on));
-	}
 	ganv_port_update_control_slider(port, value, FALSE);
 }
 
@@ -598,7 +641,7 @@ ganv_port_set_control_value_internal(GanvPort* port,
 	ganv_port_set_control_value(port, value);
 
 	// Fire signal to notify user value has changed
-	const double dvalue = value;
+	const double dvalue = port->impl->control->value;
 	g_signal_emit(port, port_signals[PORT_VALUE_CHANGED], 0, dvalue, NULL);
 }
 
@@ -645,12 +688,12 @@ ganv_port_get_natural_width(const GanvPort* port)
 	} else {
 		w = ganv_module_get_empty_port_depth(ganv_port_get_module(port));
 	}
-	if (port->impl->control && port->impl->control->label &&
-	    (GANV_ITEM(port->impl->control->label)->object.flags
+	if (port->impl->value_label &&
+	    (GANV_ITEM(port->impl->value_label)->object.flags
 	     & GANV_ITEM_VISIBLE)) {
 		double label_w;
-		g_object_get(port->impl->control->label, "width", &label_w, NULL);
-		w += (PORT_LABEL_HPAD * 4.0);
+		g_object_get(port->impl->value_label, "width", &label_w, NULL);
+		w += label_w + PORT_LABEL_HPAD;
 	}
 	return w;
 }
